/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.admin;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.MBeanNames;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jmx.api.api.AdminServiceClient;
import org.ow2.petals.jmx.api.api.DeploymentServiceClient;
import org.ow2.petals.jmx.api.api.InstallationServiceClient;
import org.ow2.petals.microkernel.api.communication.JMXService;
import org.ow2.petals.microkernel.api.container.ComponentLifeCycle;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.ServiceAssemblyLifeCycle;
import org.ow2.petals.microkernel.api.container.ServiceUnitLifeCycle;
import org.ow2.petals.microkernel.api.jbi.management.AdminService;
import org.ow2.petals.microkernel.jbi.management.MBeanNameBuilder;
import org.ow2.petals.microkernel.server.MBeanHelper;
import org.ow2.petals.microkernel.util.ParameterCheckHelper;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * TODO AdminServiceImpl : check the exceptions catch.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author wjoseph - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = AdminService.class))
public class AdminServiceImpl implements AdminService {

    /**
     * Logger wrapper
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(AdminService.COMPONENT_LOGGER_NAME));

    /**
     * The JMX Fractal service
     */
    @Requires(name = "jmx")
    private JMXService jmxService;

    @Requires(name = "mbean-name-builder")
    private MBeanNameBuilder mBeanNames;

    @Requires(name = "containerService")
    private ContainerService containerService;

    private MBeanServer mbeanServer;

    /**
     * {@inheritDoc}
     */
    public ObjectName[] getBindingComponents() {
        this.log.start();
        try {
            final List<ObjectName> result = new ArrayList<ObjectName>();
            for (final String name : this.containerService.getBindingComponentsNames()) {
                result.add(this.mBeanNames.createBindingComponentMBeanName(name));
            }
            return result.toArray(new ObjectName[0]);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public ObjectName[] getEngineComponents() {
        this.log.start();
        try {
            final List<ObjectName> result = new ArrayList<ObjectName>();
            for (final String name : this.containerService.getServiceEnginesNames()) {
                result.add(this.mBeanNames.createEngineComponentMBeanName(name));
            }
            return result.toArray(new ObjectName[0]);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public ObjectName getComponentByName(final String name) {
        this.log.call();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(name, "name cannot be null", this.log);

        final ComponentLifeCycle componentLifecycle = this.containerService.getComponentLifecycleByName(name);
        if (componentLifecycle != null) {
            if (this.containerService.isBindingComponent(name)) {
                return this.mBeanNames.createBindingComponentMBeanName(name);
            } else if (this.containerService.isServiceEngine(name)) {
                return this.mBeanNames.createEngineComponentMBeanName(name);
            } else {
                return null;
            }
        } else {
            return null;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.AdminServiceMBean#getSystemInfo()
     */
    public String getSystemInfo() {
        return new String("Petals JBI Container "
                + AdminServiceImpl.class.getPackage().getImplementationVersion() + "\n"
                + System.getProperty("java.runtime.name") + " "
                + System.getProperty("java.vm.version") + " "
                + System.getProperty("java.vm.vendor") + "\n" + System.getProperty("os.name") + " "
                + System.getProperty("os.version") + " " + System.getProperty("os.arch"));
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.jbi.management.AdminServiceMBean#getSystemService(java.lang.String)
     */
    public ObjectName getSystemService(String serviceName) {
        this.log.call();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(serviceName,
                "serviceName cannot be null", this.log);

        ObjectName result = null;
        ObjectName[] systemServices = this.getSystemServices();

        for (ObjectName objectName : systemServices) {
            if (objectName.getCanonicalName().equals(serviceName)) {
                result = objectName;
                break;
            }
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.AdminServiceMBean#getSystemServices()
     */
    public ObjectName[] getSystemServices() {
        this.log.call();

        ObjectName[] names = new ObjectName[3];

        try {
            names[0] = MBeanHelper.retrieveServiceMBean(
                    AdminServiceClient.ADMIN_SERVICE_MBEAN_NAME, this.mbeanServer);
            names[1] = MBeanHelper.retrieveServiceMBean(
                    DeploymentServiceClient.DEPLOYMENT_SERVICE_MBEAN_NAME,
                    this.mbeanServer);
            names[2] = MBeanHelper.retrieveServiceMBean(
                    InstallationServiceClient.INSTALLATION_SERVICE_MBEAN_NAME,
                    this.mbeanServer);
        } catch (MalformedObjectNameException e) {
            this.log.warning("A system service name can't be retrieved !", e);
        }

        return names;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isBinding(String componentName) {
        ParameterCheckHelper.isNullOrEmptyParameterWithLog(componentName,
                "componentName cannot be null", this.log);

        return this.containerService.isBindingComponent(componentName);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isEngine(String componentName) {
        ParameterCheckHelper.isNullOrEmptyParameterWithLog(componentName,
                "componentName cannot be null", this.log);

        return this.containerService.isServiceEngine(componentName);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isExchangeWithConsumerOkayForComponent(final String componentName,
            final ServiceEndpoint internalEndpoint, final MessageExchange exchange) throws JBIException {
        this.log.start();
        try {
            final Installer installer = this.containerService.getInstallerByName(componentName);
            if (installer == null) {
                throw new JBIException("Cannot find a component '" + componentName + "' in the container");
            }

            return installer.getComponent().isExchangeWithConsumerOkay(internalEndpoint, exchange);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isExchangeWithProviderOkayForComponent(final String componentName,
            final ServiceEndpoint internalEndpoint, final MessageExchange exchange) throws JBIException {
        this.log.start();
        try {
            final Installer installer = this.containerService.getInstallerByName(componentName);
            if (installer == null) {
                throw new JBIException("Cannot find a component '" + componentName + "' in the container");
            }

            return installer.getComponent().isExchangeWithProviderOkay(internalEndpoint,
                exchange);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public MBeanNames getMBeanNames() {
        return this.mBeanNames;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.management.admin.AdminService#getMBeanServer()
     */
    public MBeanServer getMBeanServer() {
        return this.mbeanServer;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.management.admin.AdminService#
     * getServiceUnitsLifeCyclesForComponent(java.lang.String)
     */
    public List<ServiceUnitLifeCycle> getServiceUnitsLifeCyclesForComponent(String component) {
        this.log.call();

        List<ServiceUnitLifeCycle> serviceUnitLifeCycles = new ArrayList<ServiceUnitLifeCycle>();

        for (final ServiceAssemblyLifeCycle serviceAssemblyLifeCycle : this.containerService.getServiceAssemblies()
                .values()) {
            for (ServiceUnitLifeCycle serviceUnitLifeCycle : serviceAssemblyLifeCycle.getServiceUnitLifeCycles()) {
                if (serviceUnitLifeCycle.getTargetComponentName().equals(component)) {
                    serviceUnitLifeCycles.add(serviceUnitLifeCycle);
                }
            }
        }

        return serviceUnitLifeCycles;
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        this.mbeanServer = this.jmxService.getLocalJMXServer();

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();

    }

}
