/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.exchange;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.messaging.PetalsDeliveryChannel;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.jbi.messaging.RoutingException;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.pooling.PoolException;
import com.ebmwebsourcing.easycommons.xml.SourceHelper;

/**
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
public class DeliveryChannelImpl implements PetalsDeliveryChannel {

    private final PetalsComponentContext componentContext;

    private final LoggingUtil log;

    private final Logger logger;

    protected boolean opened = true;

    /**
     * Constructor
     * 
     * @param context
     * @param router
     * @param logger
     */
    public DeliveryChannelImpl(final PetalsComponentContext context) {
        this.logger = context.getLogger();

        this.log = new LoggingUtil(this.logger);

        this.componentContext = context;
    }

    /**
     * Check if the <code>MessageExchange</code> is Ok (not null and not
     * terminated)
     * 
     * @throws MessagingException
     */
    private final void checkMessageExchange(final PetalsMessageExchange exchange)
            throws MessagingException {
        if (exchange == null) {
            throw new MessagingException("MessageExchange is null");
        }

        final String termination = exchange.getTermination();
        if (termination != null) {
            throw new MessagingException(
                    "MessageExchange is terminated, it can not be sent any more (reason: " + termination + ")");
        }
    }

    @Override
    public MessageExchangeWrapper accept() throws MessagingException {
        this.log.call();
        final MessageExchangeWrapper me = this.poll(-1);

        return me;
    }

    @Override
    public MessageExchangeWrapper accept(final long timeoutMS) throws MessagingException {
        this.log.call(timeoutMS);
        if (timeoutMS < 0) {
            throw new MessagingException("Timeout must be positive");
        }
        final MessageExchangeWrapper me = this.poll(timeoutMS);
        return me;
    }

    private void checkDeliveryChannelIsOpened() throws MessagingException {
        if (!this.opened) {
            throw new MessagingException("DeliveryChannel is closed");
        }
    }

    /**
     * This method wake up all threads that are waiting for a message. This method deregister all endpoints for this
     * component.
     */
    @Override
    public synchronized void close() throws MessagingException {
        this.log.start();

        this.checkDeliveryChannelIsOpened();

        this.opened = false;

        this.log.end();
    }

    @Override
    public MessageExchangeFactory createExchangeFactory() {
        this.log.call();
        PersistedMessageExchangeFactoryImpl exchangeFactoryImpl = PersistedMessageExchangeFactoryImpl.create(this.componentContext.getAddress(), this.logger);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final QName interfaceName) {
        this.log.call(interfaceName);
        PersistedMessageExchangeFactoryImpl exchangeFactoryImpl = PersistedMessageExchangeFactoryImpl.createForInterface(this.componentContext.getAddress(), this.logger, interfaceName);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final ServiceEndpoint endpoint) {
        this.log.call(endpoint);
        PersistedMessageExchangeFactoryImpl exchangeFactoryImpl = PersistedMessageExchangeFactoryImpl.createForEndpoint(this.componentContext.getAddress(), this.logger, endpoint);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactoryForService(final QName serviceName) {
        this.log.call(serviceName);
        PersistedMessageExchangeFactoryImpl exchangeFactoryImpl = PersistedMessageExchangeFactoryImpl.createForService(this.componentContext.getAddress(), this.logger, serviceName);
        return exchangeFactoryImpl;
    }

    @Override
    public boolean isOpened() {
        return this.opened;
    }

    @Override
    public void send(final MessageExchange exchange) throws MessagingException {
        this.log.call("Exchange Id:" + exchange.getExchangeId());

        final PetalsMessageExchange answer;
        try {
            answer = this.sendExchange(((MessageExchangeWrapper) exchange).getMessageExchange(), 0, false);
        } catch (InterruptedException e) {
            final String message = "Asynchronous send interrupted, this is impossible!";
            this.log.error(message, e);
            this.log.assertOrLog(false, message);
            throw new MessagingException(message, e);
        }
        // it can't be otherwise, we are not in sendSync
        assert answer == null;
    }

    @Override
    public boolean sendSync(final MessageExchange exchange) throws MessagingException {
        return this.sendSync(exchange, 0);
    }

    /**
     * Note: the timeout count actually starts when the Router take the lead (once
     * {@link RouterService#sendSync(PetalsComponentContext, PetalsMessageExchange, long)} has been called).
     */
    @Override
    public boolean sendSync(final MessageExchange exchange, final long timeoutMS)
            throws MessagingException {
        this.log.call("Exchange Id:" + exchange.getExchangeId() + " - Timeout:" + timeoutMS);

        if (timeoutMS < 0) {
            throw new IllegalArgumentException("Timeout must be superior or equal to 0.");
        }
        
        // the exchange can only be a MessageExchangeWrapper
        final MessageExchangeWrapper exchangeDecorator = (MessageExchangeWrapper) exchange;
        final PetalsMessageExchange messageExchange = exchangeDecorator.getMessageExchange();

        PetalsMessageExchange answer = null;
        try {
            // the answer will be returned
            answer = this.sendExchange(messageExchange, timeoutMS, true);
        } catch (final MessagingException e) {
            this.log.debug("Exception while sending: we set the message in ERROR");
            messageExchange.setError(e);
            // it's terminated: this error won't be sent again by the caller of this method
            messageExchange.setTermination("Error during sendSync()");
            throw e;
        } catch (final InterruptedException e) {
            // Even if one of the causes is an InterruptedException,
            // we must throw the MessagingException according to the specification
            this.log.debug(
                    "sendSync() was interrupted, converting to a MessagingException as defined in the JBI specification.");
            messageExchange.setStatus(ExchangeStatus.ERROR);
            messageExchange.setTermination("Interruption during sendSync()");
            throw new MessagingException(e);
        }

        if (answer != null) {
            exchangeDecorator.setMessageExchange(answer);
        } else {
            // we timed-out
            messageExchange.setStatus(ExchangeStatus.ERROR);
            // so that we are sure nobody will try to send it again, we terminate the exchange
            messageExchange.setTermination("Timeout");
        }

        this.log.end();

        return answer != null;
    }

    /**
     * Log the current message, it can be an Out, Fault or In message.
     * 
     * @param exchange a message exchange decorator
     * @throws JBIException
     * @throws TransformerException
     *             if an unrecoverable error occurs during the course of the
     *             transformation
     * @throws PoolException
     *             This method uses a transformer got from a transformer pool.
     *             This unchecked exception is thrown if the current thread
     *             waits for a transformer of the pool and is interrupted when
     *             the pool policy is WAIT or if there is no more available
     *             transformer in the transformer pool when the pool policy is
     *             REJECT
     */
    private void logMessage(final PetalsMessageExchange exchange) {
        java.util.logging.Logger componentLogger = null;
        // dirty but will log to the component logger.
        try {
            componentLogger = this.componentContext.getLogger("", null);
            if (componentLogger.isLoggable(Level.FINEST)) {
                if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                    NormalizedMessage normalizedMessage = exchange.getMessage(PetalsMessageExchange.OUT_MSG);
                    String messageType = PetalsMessageExchange.OUT_MSG;
                    if (normalizedMessage == null) {
                        normalizedMessage = exchange.getFault();
                        messageType = "Fault";
                    }
                    if (normalizedMessage == null) {
                        normalizedMessage = exchange.getMessage(PetalsMessageExchange.IN_MSG);
                        messageType = PetalsMessageExchange.IN_MSG;
                    }

                    componentLogger.finest("Message " + messageType + " with Id '"
                            + exchange.getExchangeId() + "' to send");
                    componentLogger.finest("Operation : " + exchange.getOperation());
                    componentLogger.finest("MEP : " + exchange.getPattern());

                    for (final Object propertyName : exchange.getPropertyNames()) {
                        componentLogger.finest("Exchange property '" + propertyName + "' : "
                                + exchange.getProperty((String) propertyName));
                    }

                    for (final Object propertyName : normalizedMessage.getPropertyNames()) {
                        componentLogger.finest("Message property '" + propertyName + "' : "
                                + normalizedMessage.getProperty((String) propertyName));
                    }

                    Source content = normalizedMessage.getContent();
                    if (content != null) {
                        componentLogger.finest("Message Content : " + SourceHelper.toString(content));
                    } else {
                        componentLogger.finest("Message Content is null");
                    }

                    for (final Object attachmentName : normalizedMessage.getAttachmentNames()) {
                        componentLogger.finest("Got a message attachment named '" + attachmentName
                                + "'");
                    }
                } else if (ExchangeStatus.DONE.equals(exchange.getStatus())) {
                    componentLogger.finest("Message acknowledgement DONE with Id '"
                            + exchange.getExchangeId() + "' to send");
                } else {
                    componentLogger.finest("Message acknowledgement ERROR with Id '"
                            + exchange.getExchangeId() + "' to send");
                    componentLogger.finest("Message Error : " + exchange.getError().getMessage());
                }
            }
        } catch (final Exception e) {
            if (componentLogger != null) {
                componentLogger.fine("Failed to log the content of the message to send: "
                        + e.getMessage());
            }
        }
    }

    /**
     * Poll for a receipt message exchange.
     * 
     * @param timeoutMS
     * 
     * @return The received message exchange, or {@code null} if timeout
     *         occurred
     * 
     * @throws MessagingException
     */
    private MessageExchangeWrapper poll(final long timeoutMS) throws MessagingException {
        this.log.start();

        this.checkDeliveryChannelIsOpened();

        final PetalsMessageExchange exchange;
        try {
            exchange = this.componentContext.getRouter().receive(this.componentContext,
                    timeoutMS);
        } catch (final RoutingException e) {
            throw new MessagingException(e);
        } catch (final InterruptedException e) {
            // Even if one of the causes is an InterruptedException,
            // we must throw the MessagingException according to the specification
            this.log.debug(
                    "accept() was interrupted, converting to a MessagingException as defined in the JBI specification.");
            throw new MessagingException("accept() was interrupted", e);
        }

        final MessageExchangeWrapper exchangeDecorator;
        if (exchange != null) {
            // TODO make it so that we can use the factory??
            // the idea is that we are giving the exchange to its owner, so the observer role is the role of the
            // exchange.
            exchangeDecorator = new MessageExchangeWrapper(exchange);
        } else {
            exchangeDecorator = null;
        }

        this.log.end();

        return exchangeDecorator;
    }

    /**
     * Send the exchange to the underlying layer.
     * 
     * If the message to exchange is a mere status information, and the exchange is set to not send acknowledgment, the
     * message is not sent.
     * 
     * @throws InterruptedException
     */
    private PetalsMessageExchange sendExchange(final PetalsMessageExchange exchange, final long timeout,
            final boolean sync) throws MessagingException, InterruptedException {
        this.checkDeliveryChannelIsOpened();

        this.checkMessageExchange(exchange);

        this.logMessage(exchange);

        // switch the classloader to the container one
        final ClassLoader componentClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        try {
            // check if the exchange is terminated and mark it as such.
            // The idea is that an exchange is terminated once it has been sent by the last actor of the exchange
            if (!ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                exchange.setTermination("Pattern is finished, ending status sent");
                if (sync) {
                    throw new MessagingException(
                            "Sending an acknowledgment in synchronous mode is forbidden");
                }
            }

            if (sync) {
                final PetalsMessageExchange answer = this.componentContext.getRouter().sendSync(this.componentContext,
                        exchange, timeout);

                this.log.end();

                return answer;
            } else {
                this.componentContext.getRouter().send(this.componentContext, exchange);

                this.log.end();

                return null;
            }

        } catch (final JBIException e) {
            throw new MessagingException(e);
        } finally {
            Thread.currentThread().setContextClassLoader(componentClassLoader);
        }
    }
}
