/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.routing;

import static javax.jbi.management.LifeCycleMBean.SHUTDOWN;
import static javax.jbi.management.LifeCycleMBean.STARTED;
import static javax.jbi.management.LifeCycleMBean.STOPPED;
import static javax.jbi.messaging.MessageExchange.Role.CONSUMER;
import static javax.jbi.messaging.MessageExchange.Role.PROVIDER;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.jbi.messaging.MessageExchange.Role;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Cardinality;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.container.ServiceUnitLifeCycle;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.jbi.messaging.RoutingException;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.InstallModule;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.ReceiverModule;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.SenderModule;
import org.ow2.petals.microkernel.jbi.messaging.routing.monitoring.RouterMonitorService;
import org.ow2.petals.microkernel.transport.TransportException;
import org.ow2.petals.microkernel.transport.TransportListener;
import org.ow2.petals.microkernel.transport.Transporter;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * Routes messages to their destinations. 
 * <p>
 * Performs an itinerary resolution,
 * adding for example a transformation service before delivering the message to
 * its recipient, and perform an address resolution, finding the endpoint
 * satisfying the the destination information (finding an endpoint for a
 * specified service). Resolutions are based on defined rules.
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
@Component(provides = {
 @Interface(name = "service", signature = RouterService.class),
        @Interface(name = "transportlistener", signature = TransportListener.class) })
public class RouterServiceImpl implements RouterService, TransportListener {

    private static final String INSTALLMODULE_FRACTAL_PREFIX = "installmodule-";

    private static final String PROVIDER_SUFFIX = "-provider";

    private static final String RECEIVERMODULE_FRACTAL_PREFIX = "receivermodule-";

    private static final String SENDERMODULE_FRACTAL_PREFIX = "sendermodule-";

    // private static final String CONSUMER_SUFFIX = "-consumer";

    /**
     * The timeout of the stop of the traffic
     */
    private static final int STOP_TRAFFIC_TIMEOUT = 10000;

    private static final String TRANSPORTER_FRACTAL_PREFIX = "transporter-";

    /**
     * The map of the exchanges queues. One queue per installed component.
     */
    private Map<String, BlockingQueue<PetalsMessageExchange>> exchangeQueues;

    /**
     * Configuration Service
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The Router Monitor Fractal component
     */
    @Requires(name = "routermonitor")
    private RouterMonitorService routerMonitorService;

    /**
     * The list of installModule Fractal components
     */
    @Requires(name = INSTALLMODULE_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, InstallModule> installModules = new HashMap<String, InstallModule>();

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(RouterService.COMPONENT_LOGGER_NAME));

    /**
     * The map of the pending exchanges. One List per Provides of SU stopped or shut down.
     * 
     * Messages are stored there until the SU is started
     */
    private ConcurrentMap<String, List<PetalsMessageExchange>> pendingMessageExchanges;

    /**
     * The list of ReceiverModule Fractal components
     * 
     * Note: the order of the modules is important, hence the use of a TreeMap and keys that are ordered
     * 
     * TODO is it important for this one? not so sure...
     */
    @Requires(name = RECEIVERMODULE_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, ReceiverModule> receiverModules = new TreeMap<String, ReceiverModule>();

    /**
     * The list of SenderModule Fractal components
     * 
     * Note: the order of the modules is important, hence the use of a TreeMap and keys that are ordered
     * 
     * TODO is it important for this one? not so sure...
     */
    @Requires(name = SENDERMODULE_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, SenderModule> senderModules = new TreeMap<String, SenderModule>();

    /**
     * Traffic flag to handle stop of traffic
     * 
     * It is volatile to ensure that every thread has its real value
     */
    private volatile boolean stopTraffic;

    /**
     * The set of the threads that are blocking on a poll
     */
    private Set<Thread> threadsList;

    /**
     * the Transporter Fractal components
     */
    @Requires(name = TRANSPORTER_FRACTAL_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, Transporter> transporters = new HashMap<String, Transporter>();

    /**
     * Offering timeout used by component queues
     */
    long transporterQueueOfferingTimeout;

    @Override
    public void addComponent(final PetalsComponentContext componentContext) throws RoutingException {
        this.log.start();

        for (final InstallModule installModule : this.installModules.values()) {
            installModule.addComponent(componentContext);
        }

        final ContainerConfiguration containerConfiguration = this.configurationService
                .getContainerConfiguration();

        final int transporterQueueMaxSize = containerConfiguration.getTransporterQueueMaxSize();
        if (this.log.isDebugEnabled()) {
            this.log.debug("Using a queue bounded to " + transporterQueueMaxSize
                    + " for component '" + componentContext.getComponentName() + "'.");
        }
        this.exchangeQueues.put(componentContext.getComponentName(),
                new ArrayBlockingQueue<PetalsMessageExchange>(transporterQueueMaxSize));

        this.log.end();
    }

    @Override
    public void exchangeSent(PetalsMessageExchange exchange) {
        this.routerMonitorService.exchangeSent(exchange);
    }

    @Override
    public void modifiedSALifeCycle(final List<ServiceUnitLifeCycle> serviceUnitLifeCycles) {
        this.log.start();

        for (final ServiceUnitLifeCycle serviceUnitLifeCycle : serviceUnitLifeCycles) {
            final List<org.ow2.petals.jbi.descriptor.original.generated.Provides> providesList = serviceUnitLifeCycle
                    .getServiceUnitDescriptor().getServices().getProvides();
            if (providesList != null) {
                final String suState = serviceUnitLifeCycle.getCurrentState();
                for (final org.ow2.petals.jbi.descriptor.original.generated.Provides provides : providesList) {
                    final String uniqueId = provides.getEndpointName() + "-" + provides.getServiceName()
                            + PROVIDER_SUFFIX;
                    if (STOPPED.equals(suState) || SHUTDOWN.equals(suState)) {
                        // The SU is not started, so we will start storing exchanges before it is
                        // Note: this relies on the fact that the state is set to SHUTDOWN when an SU is initialised
                        // use a LinkedList since elements are either added or retrieved all at once
                        this.pendingMessageExchanges.putIfAbsent(uniqueId, new LinkedList<PetalsMessageExchange>());
                    } else if (STARTED.equals(suState)) {
                        // The SU is now started, let's get the waiting exchange in order to process them
                        final BlockingQueue<PetalsMessageExchange> componentQueue = this.exchangeQueues
                                .get(serviceUnitLifeCycle.getTargetComponentName());
                        // synchronisation to avoid the remove operation to happen between
                        // the retrieval of the list and the adding of the exchange (in onExchange)
                        synchronized (this.pendingMessageExchanges) {
                            componentQueue.addAll(this.pendingMessageExchanges.remove(uniqueId));
                        }
                    }
                }
            }
            // TODO: find a way to handle consumes when they are shut down
        }

        this.log.end();
    }

    @Override
    public void onExchange(final PetalsMessageExchange exchange) {
        this.log.start();

        String componentName = null;

        this.routerMonitorService.exchangeReceived(exchange);

        if (exchange.getRole().equals(CONSUMER)) {
            // TODO: handle the List of shut down consumes
            componentName = ((PetalsServiceEndpoint) exchange.getConsumerEndpoint()).getLocation().getComponentName();
        } else if (exchange.getRole().equals(PROVIDER)) {
            final PetalsServiceEndpoint endpoint = (PetalsServiceEndpoint) exchange.getEndpoint();
            final String uniqueId = endpoint.getEndpointName() + "-" + endpoint.getServiceName() + PROVIDER_SUFFIX;
            // synchronisation so that the list is not removed from the map
            // inbetween the get and the add!
            synchronized (this.pendingMessageExchanges) {
                final List<PetalsMessageExchange> pme = this.pendingMessageExchanges.get(uniqueId);
                if (pme != null) {
                    this.log.debug("SU not started, store the exchange");
                    pme.add(exchange);
                } else {
                    componentName = ((PetalsServiceEndpoint) exchange.getEndpoint()).getLocation().getComponentName();
                }
            }
        }

        if (componentName != null) {
            try {
                this.exchangeQueues.get(componentName).offer(exchange,
                        this.transporterQueueOfferingTimeout, TimeUnit.MILLISECONDS);
            } catch (final InterruptedException e) {
                this.log.error("A message exchange has been lost because of exhaustion of the component transporter queue.");
            }
        }

        this.log.end();
    }

    @Override
    public PetalsMessageExchange receive(final PetalsComponentContext componentContext, final long timeoutMS)
            throws RoutingException, InterruptedException {

        final String componentName = componentContext.getComponentName();

        this.log.start("Component : " + componentName + " - Timeout : " + timeoutMS);

        this.checkStopTraffic();

        final PetalsMessageExchange exchange;
        if (timeoutMS == 0) {
            // non-blocking call
            exchange = this.exchangeQueues.get(componentName).poll();
        } else {
            this.threadsList.add(Thread.currentThread());
            try {
                if (timeoutMS > 0) {
                    // blocking call with timeout
                    exchange = this.exchangeQueues.get(componentName).poll(timeoutMS, TimeUnit.MILLISECONDS);
                } else {
                    // blocking call without timeout
                    exchange = this.exchangeQueues.get(componentName).take();
                }
            } finally {
                this.threadsList.remove(Thread.currentThread());
            }
        }

        if (exchange != null) {
            for (final ReceiverModule receiverModule : this.receiverModules.values()) {
                receiverModule.receiveExchange(exchange, componentContext);
            }
            this.log.end("Exchange Id : " + exchange.getExchangeId() + " - Component : " + componentName);
        } else {
            this.log.end("No exchange - Component : " + componentName);
        }

        return exchange;
    }

    @Override
    public void removeComponent(final PetalsComponentContext componentContext) throws RoutingException {
        this.log.start();

        for (final InstallModule installModule : this.installModules.values()) {
            installModule.removeComponent(componentContext);
        }

        this.exchangeQueues.remove(componentContext.getComponentName());

        this.log.end();
    }

    @Override
    public void send(final PetalsComponentContext source, final PetalsMessageExchange exchange)
            throws RoutingException {
        this.log.call();

        this.checkStopTraffic();

        // Get the list of elected endpoints from the sender modules
        // using a LinkedHashMap to keep the insertion order
        final Map<PetalsServiceEndpoint, TransportSendContext> electedDestinations = new LinkedHashMap<PetalsServiceEndpoint, TransportSendContext>();
        for (final SenderModule senderModule : this.senderModules.values()) {
            senderModule.electEndpoints(electedDestinations, source, exchange);
        }

        try {
            if (CONSUMER.equals(exchange.getRole())) {
                exchange.setRole(PROVIDER);
            } else {
                exchange.setRole(CONSUMER);
                // only one destination for a consumer
                this.log.assertOrLog(electedDestinations.values().size() == 1,
                        "There shouldn't be multiple elected destination if an exchange is sent back to a Consumer");
            }
            final PetalsMessageExchange answer = this.doSend(electedDestinations, exchange, false);
            // it can't be otherwise, we are not in sendSync
            assert answer == null;
        } catch (RoutingException e) {
            // reset the roles
            if (CONSUMER.equals(exchange.getRole())) {
                exchange.setRole(PROVIDER);
            } else {
                exchange.setRole(CONSUMER);
            }
            throw e;
        } catch (final InterruptedException e) {
            final String message = "Asynchronous send interrupted, this is impossible!";
            this.log.error(message, e);
            this.log.assertOrLog(false, message);
            throw new RoutingException(message, e);
        }
    }

    @Override
    public PetalsMessageExchange sendSync(final PetalsComponentContext source, final PetalsMessageExchange exchange,
            final long timeout) throws RoutingException, InterruptedException {
        this.log.start();

        this.checkStopTraffic();

        final Long deadline;
        if (timeout > 0) {
            deadline = System.currentTimeMillis() + timeout;
            exchange.setProperty(PROPERTY_ROUTER_TIMEOUT_DEADLINE, deadline);
        } else {
            deadline = null;
        }

        PetalsMessageExchange answer = null;
        final Role originalRole = exchange.getRole();

        // TODO factorise with send, there is so much duplicated code...
        try {
            // Get the list of elected endpoints from the sender modules
            // using a LinkedHashMap to keep the insertion order
            final Map<PetalsServiceEndpoint, TransportSendContext> electedDestinations = new LinkedHashMap<PetalsServiceEndpoint, TransportSendContext>();
            for (final SenderModule senderModule : this.senderModules.values()) {
                senderModule.electEndpoints(electedDestinations, source, exchange);
            }

            // this could happen if a SenderModule decides to slow down the execution (see the FlowController and
            // stuffs)
            if (deadline == null || deadline >= System.currentTimeMillis()) {
                if (CONSUMER.equals(exchange.getRole())) {
                    exchange.setRole(PROVIDER);
                } else {
                    exchange.setRole(CONSUMER);
                    // only one destination for a consumer
                    this.log.assertOrLog(electedDestinations.values().size() == 1,
                            "There shouldn't be multiple elected destination if an exchange is sent back to a Consumer");
                }
                answer = this.doSend(electedDestinations, exchange, true);

                if (answer != null) {
                    this.routerMonitorService.exchangeReceived(answer);
                    for (final ReceiverModule receiverModule : this.receiverModules.values()) {
                        receiverModule.receiveExchange(answer, source);
                    }
                }
            } else {
                answer = null;
            }
        } finally {
            // then when its finished, we can put the roles back because we got the answer
            if (answer != null) {
                // we don't revert the original one, because it should be replaced and never used anymore
                answer.setRole(originalRole);
                answer.setProperty(PROPERTY_ROUTER_TIMEOUT_DEADLINE, null);
            } else {
                // we do NOT revert the role if we didn't get an answer: the consumer should is not expected to modify
                // or read the exchange now that it timed out! Moreover if we are in local mode and the provider was
                // slow, he still has the ownership of the exchange!
                exchange.setProperty(PROPERTY_ROUTER_TIMEOUT_DEADLINE, null);
            }
        }

        this.log.end();

        return answer;
    }

    @Override
    public void stopTraffic() {
        this.log.start();

        // Wait that all the pending exchanges are received or the timeout is
        // reached
        final long timeout = STOP_TRAFFIC_TIMEOUT + System.currentTimeMillis();
        boolean done = false;
        try {
            while (System.currentTimeMillis() < timeout && !done) {
                for (final BlockingQueue<PetalsMessageExchange> queue : this.exchangeQueues.values()) {
                    if (!queue.isEmpty()) {
                        // sleep for 1 second
                        Thread.sleep(1000);
                        break;
                    }
                }
                done = true;
            }
        } catch (final InterruptedException e) {
            // do nothing
        }

        if (!done) {
            this.log.warning("The timeout for stopping the Router traffic has been reached."
                    + " Some message have not beed accepted by the component, they are lost.");
        }

        this.stopTraffic = true;

        // Interrupt all the pending receive threads
        // note: noone should be able to modify the set after stopTraffic has been set to true
        // but we are not completely sure because there can't be an interleaving of threads...
        while (!this.threadsList.isEmpty()) {
            final Iterator<Thread> iterator = this.threadsList.iterator();
            while (iterator.hasNext()) {
                final Thread t = iterator.next();
                if (t.isAlive()) {
                    t.interrupt();
                }
                iterator.remove();
            }
        }

        this.log.end();
    }

    /**
     * start the router.
     * 
     * @throws RoutingException
     *             impossible to start the router
     */
    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        this.exchangeQueues = new ConcurrentHashMap<String, BlockingQueue<PetalsMessageExchange>>();
        this.pendingMessageExchanges = new ConcurrentHashMap<String, List<PetalsMessageExchange>>();
        this.threadsList = Collections.newSetFromMap(new ConcurrentHashMap<Thread, Boolean>(100));

        final ContainerConfiguration containerConfiguration = this.configurationService
                .getContainerConfiguration();
        this.transporterQueueOfferingTimeout = containerConfiguration
                .getTransporterQueueOfferingTimeout();
        if (this.log.isDebugEnabled()) {
            this.log.debug("The offering timeout of component queues is set to "
                    + transporterQueueOfferingTimeout + "'.");
        }

        this.log.end();
    }

    /**
     * stop the router.
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    /**
     * Check if the traffic is off. If so, raise an exception with the
     * interrupted flag set.
     * 
     * @throws TransportException
     */
    private void checkStopTraffic() throws RoutingException {
        if (this.stopTraffic) {
            Thread.currentThread().interrupt();
            throw new RoutingException("The Router traffic is stopped", new InterruptedException());
        }
    }

    /**
     * Sends the specified message to the provider of the service.
     * 
     * TODO the timeout should be managed in coordination with the transporter!
     * 
     * @param source
     *            the source
     * @param exchange
     *            the exchange decorator
     * @param sync
     *            {@code true} is the send is synchronous
     * @param timeout
     *            timeout of the message (in ms). 0 for no timeout one
     * @throws RoutingException
     */
    private PetalsMessageExchange doSend(final Map<PetalsServiceEndpoint, TransportSendContext> endpointDestinations,
            final PetalsMessageExchange exchange, final boolean sync) throws RoutingException, InterruptedException {
        this.log.start();

        final Long deadline = (Long) exchange.getProperty(RouterService.PROPERTY_ROUTER_TIMEOUT_DEADLINE);

        boolean retry = true;
        int attemptDelay = 0;

        // TODO improve that code, it is terribly hard to read, and factorise it with sendToConsumer
        while (retry) {
            if (attemptDelay != 0) {
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Wait " + attemptDelay + " millisecond before the next send attempt");
                }
                try {
                    Thread.sleep(attemptDelay);
                } catch (final InterruptedException e) {
                    throw new RoutingException(e);
                }
            }
            retry = false;
            final Iterator<Entry<PetalsServiceEndpoint, TransportSendContext>> iterator = endpointDestinations.entrySet()
                    .iterator();

            while (iterator.hasNext()) {
                final Entry<PetalsServiceEndpoint, TransportSendContext> entry = iterator.next();
                final PetalsServiceEndpoint electedEndpoint = entry.getKey();
                final TransportSendContext transportSendContext = entry.getValue();
                if (transportSendContext.attempt > 0) {
                    if (this.log.isDebugEnabled()) {
                        this.log.debug("Send attempt to endpoint '" + electedEndpoint.getEndpointName() + "'");
                    }

                    // if we are sending back to the consumer, then this shouldn't be touched
                    if (Role.PROVIDER.equals(exchange.getRole())) {
                        exchange.setEndpoint(electedEndpoint);
                        exchange.setService(electedEndpoint.getServiceName());
                        // the interface list contains 1 interface
                        exchange.setInterfaceName(electedEndpoint.getInterfaces()[0]);
                    }

                    try {
                        final Transporter transporter = this.transporters.get(transportSendContext.transport);
                        if (sync) {
                            if (deadline != null && deadline < System.currentTimeMillis()) {
                                this.log.end("Timeout reached!");
                                return null;
                            }
                            final PetalsMessageExchange answer = transporter.sendSync(exchange, transportSendContext);

                            this.log.end("Transporter's sendSync succeeded");

                            // if it worked let's return the answer
                            return answer;
                        } else {
                            transporter.send(exchange, transportSendContext);

                            this.log.end("Transporter's send succeeded");

                            // if it worked let's return null (it is not sendsync)
                            return null;
                        }
                    } catch (final TransportException e) {
                        transportSendContext.attempt -= 1;
                        if (transportSendContext.attempt > 0) {
                            retry = true;
                            // get the max delay
                            if (transportSendContext.delay > attemptDelay) {
                                attemptDelay = transportSendContext.delay;
                            }
                        }

                        if (iterator.hasNext() || retry) {
                            this.log.warning("The send attempt to the endpoint '" + electedEndpoint.getEndpointName()
                                    + "' with destination " + transportSendContext.destination + " failed", e);
                        } else {
                            throw new RoutingException(e);
                        }
                    }
                }
            }
        }

        this.log.end("deadcode...");

        return null;
    }
}
