/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.routing.module;

import java.io.File;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Logger;

import javax.jbi.messaging.MessageExchange.Role;
import javax.xml.namespace.QName;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.messaging.RoutingException;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.flow.PriorityModifier;
import org.ow2.petals.microkernel.jbi.messaging.routing.monitoring.RouterMonitorFilter;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;
import org.ow2.petals.regulation.RegulationBuilder;
import org.ow2.petals.regulation.RegulationException;
import org.ow2.petals.regulation.generated.Flow;
import org.ow2.petals.regulation.generated.Regulation;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The Flow Filter Module.
 * <p>
 * The configured flows to filter are monitored by this module. Each time a
 * component attempts a send, it is countered as an effective send.
 * </p>
 * <p>
 * If the number of the component send attempts overflows the maximum accepted
 * in a window, the messages are pending until the next window, by using the
 * Priority Queues.
 * </p>
 * 
 * @author Roland Naudin - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = SenderModule.class))
public class FlowFilterModule implements SenderModule {

    /**
     * The period to update the exchange priority.
     */
    private static final long PRIORITY_UPDATE_PERIOD = 1000;

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(
            Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME_FLOW_FILTER));
    /**
     * The Priority Modifier fractal component
     */
    @Requires(name = "prioritymodifier")
    private PriorityModifier priorityModifier;

    private Map<RouterMonitorFilter, _Filter> filters;

    private Timer filterTimer;

    /**
     * Inner class to update the priority of pending exchanges.
     * 
     * @author rnaudin
     * 
     */
    private class PriorityUpdater extends TimerTask {

        @Override
        public void run() {
            for (_Filter filter : FlowFilterModule.this.filters.values()) {
                synchronized (filter) {
                    filter.currentWindow++;
                    if (filter.currentWindow == filter.expectedWindow) {
                        filter.currentWindow = 0;
                        filter.messagesSent = 0;
                        while (!filter.pendingExchanges.isEmpty() && filter.messagesSent < filter.expectedFilter) {
                            final PetalsMessageExchange exchange = filter.pendingExchanges.poll();
                            try {
                                FlowFilterModule.this.priorityModifier.modifyPriority(exchange, (short) 3);
                                FlowFilterModule.this.log.debug(
                                        "Reassign priority for exchange with Id '" + exchange.getExchangeId() + "'");
                                filter.messagesSent++;
                            } catch (RoutingException e) {
                                FlowFilterModule.this.log
                                        .warning(
                                                "Failed to reassign a priority to the exchange with Id '"
                                                        + exchange.getExchangeId() + "', maybe a timeout has occurred",
                                                e);
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Inner structure.
     * 
     * @author rnaudin
     * 
     */
    private class _Filter {

        public int expectedFilter;

        public int messagesSent;

        /**
         * unbound queue
         */
        public Queue<PetalsMessageExchange> pendingExchanges;

        public int expectedWindow;

        public int currentWindow;

        public _Filter(int expectedFilter, int expectedWindow) {
            this.expectedFilter = expectedFilter;
            this.expectedWindow = expectedWindow;
            this.pendingExchanges = new ConcurrentLinkedQueue<PetalsMessageExchange>();
        }
    }

    @Override
    public void electEndpoints(final Map<PetalsServiceEndpoint, TransportSendContext> electedEndpoints,
            final PetalsComponentContext sourceComponentContext, final PetalsMessageExchange exchange)
                    throws RoutingException {

        if (Role.CONSUMER.equals(exchange.getRole()) && exchange.getMessage(PetalsMessageExchange.IN_MSG) != null) {
            for (final Entry<RouterMonitorFilter, _Filter> entry : FlowFilterModule.this.filters.entrySet()) {
                final RouterMonitorFilter routerMonitorFilter = entry.getKey();
                String endpointName = null;
                QName interfaceName = exchange.getInterfaceName();
                QName serviceName = exchange.getService();
                QName operationName = exchange.getOperation();

                if (exchange.getEndpoint() != null) {
                    endpointName = exchange.getEndpoint().getEndpointName();
                    interfaceName = exchange.getEndpoint().getInterfaces()[0];
                    serviceName = exchange.getEndpoint().getServiceName();
                }

                if ((routerMonitorFilter.getInterfaceName() == null || (interfaceName != null && interfaceName
                        .equals(routerMonitorFilter.getInterfaceName())))
                        && (routerMonitorFilter.getServiceName() == null || (serviceName != null && serviceName
                                .equals(routerMonitorFilter.getServiceName())))
                        && (routerMonitorFilter.getEndpointName() == null || (endpointName != null && endpointName
                                .equals(routerMonitorFilter.getEndpointName())))
                        && (routerMonitorFilter.getOperationName() == null || (operationName != null && operationName
                                .equals(routerMonitorFilter.getOperationName())))) {
                    final _Filter filter = entry.getValue();
                    boolean toFilter = false;
                    synchronized (filter) {
                        if (filter.messagesSent == filter.expectedFilter) {
                            filter.pendingExchanges.add(exchange);
                            toFilter = true;
                        } else {
                            filter.messagesSent++;
                        }
                    }

                    if (toFilter) {
                        exchange.setProperty(PriorityOrdererModule.PROPERTY_ROUTER_PRIORITY, "0");
                        this.log.debug("Filter the exchange with Id '" + exchange.getExchangeId() + "'");
                    }

                }
            }
        }
    }

    /**
     * Start the Fractal component
     * 
     * @throws RegulationException
     * 
     * @throws RoutingException
     */
    @Lifecycle(step = Step.START)
    public void start() throws RegulationException {
        this.log.start();

        this.filters = new HashMap<RouterMonitorFilter, _Filter>();
        this.filterTimer = new Timer("Router - Flow filter module - Filter");

        try {
            if (this.loadRegulationConfiguration()) {
                this.filterTimer.schedule(new PriorityUpdater(), PRIORITY_UPDATE_PERIOD,
                        PRIORITY_UPDATE_PERIOD);
            }
        } catch (RegulationException e) {
            this.log.error("Failed to start Priority Regulator service", e);
            throw e;
        }

        this.log.end();
    }

    /**
     * Stop the Fractal component
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();

        this.filterTimer.cancel();
    }

    /**
     * load the Regulation configuration.
     * 
     * @throws RegulationException
     */
    private boolean loadRegulationConfiguration() throws RegulationException {
        boolean result = false;
        final URL regulationUrl = this.getClass().getResource(
                "/" + FlowControllerModule.REGULATION_CONFIGURATION_FILE);

        if (regulationUrl != null) {
            final Regulation regulation;
            try {
                regulation = RegulationBuilder.createRegulation(new File(regulationUrl.toURI()));
            } catch (final URISyntaxException e) {
                throw new RegulationException("Invalid URL", e);
            }
            for (Flow flow : regulation.getFlow()) {
                RouterMonitorFilter routerMonitorFilter = new RouterMonitorFilter(flow
                        .getInterface(), flow.getService(), flow.getEndpoint(),
                        flow.getOperation(), false);
                if (flow.getFilter() != null) {
                    this.filters.put(routerMonitorFilter, new _Filter(flow.getFilter().getValue(),
                            flow.getFilter().getWindow()));
                    this.log.info("Add a flow filter on " + routerMonitorFilter
                            + " with threshold " + flow.getFilter().getValue() + " on window "
                            + flow.getFilter().getWindow());
                }
            }
            result = true;
        }

        return result;
    }
}
