/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.routing.monitoring;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jmx.commons.PetalsMonitoringServiceConstants.Exchange;
import org.ow2.petals.jmx.commons.PetalsMonitoringServiceConstants.MEP;
import org.ow2.petals.jmx.commons.PetalsMonitoringServiceConstants.NormalizedMessage;
import org.ow2.petals.jmx.commons.PetalsMonitoringServiceConstants.Role;
import org.ow2.petals.jmx.commons.PetalsMonitoringServiceConstants.RouterMonitor;
import org.ow2.petals.jmx.commons.PetalsMonitoringServiceConstants.Status;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.messaging.RoutingException;
import org.ow2.petals.microkernel.api.jbi.messaging.exchange.PersistedPetalsMessageExchange;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.api.system.persistence.PersistenceService;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.SenderModule;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.pooling.PoolException;
import com.ebmwebsourcing.easycommons.xml.SourceHelper;

/**
 * The NMR monitor service.
 * @author Roland Naudin - EBM WebSourcing
 */
@Component(provides = { @Interface(name = "service", signature = RouterMonitorService.class),
        @Interface(name = "sendermodule", signature = SenderModule.class) })
public class RouterMonitorServiceImpl implements RouterMonitorService, SenderModule {

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));

    /**
     * The Persistence Fractal Service
     */
    @Requires(name = "persistence")
    private PersistenceService persistenceService;

    /**
     * The Router Monitor filters.
     */
    private Set<RouterMonitorFilter> routerMonitorFilters;

    @Override
    public void clearMonitorStorage() {
        this.persistenceService.clearExchangeStorage();
    }

    @Override
    public void electEndpoints(final Map<PetalsServiceEndpoint, TransportSendContext> electedEndpoints,
            final PetalsComponentContext sourceComponentContext, final PetalsMessageExchange exchange)
                    throws RoutingException {
        this.log.call();

        if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())
                && exchange.getMessage(PetalsMessageExchange.IN_MSG) != null) {
            // this method only makes sense for a message that can be persisted!
            if (!(exchange instanceof PersistedPetalsMessageExchange)) {
                this.log.warning(
                        "The message is not a PersistedPetalsMessageExchange: " + exchange.getClass().getName());
                return;
            }
            checkMonitored((PersistedPetalsMessageExchange) exchange);
        }
    }

    @Override
    public void exchangeReceived(final PetalsMessageExchange exchange) {
        this.log.call("Exchange Id : " + exchange.getExchangeId());

        // this method only makes sense for a message that can be persisted!
        if (!(exchange instanceof PersistedPetalsMessageExchange)) {
            this.log.warning(
"The message is not a PersistedPetalsMessageExchange: " + exchange.getClass().getName());
            return;
        }

        final PersistedPetalsMessageExchange pmex = (PersistedPetalsMessageExchange) exchange;

        try {
            if (pmex.isMonitored() || pmex.isPersisted()) {
                if (((PetalsServiceEndpoint) pmex.getConsumerEndpoint()).getLocation().getContainerName()
                        .equals(((PetalsServiceEndpoint) pmex.getEndpoint()).getLocation().getContainerName())) {
                    // local send, update just the receipt timestamp
                    this.persistenceService.updateMessageExchange(pmex, false, Boolean.FALSE);
                } else {
                    if (MessageExchange.Role.PROVIDER.equals(pmex.getRole())
                            && pmex.getMessage(PetalsMessageExchange.IN_MSG) != null) {
                        // first time the exchange is add
                        this.persistenceService.addMessageExchange(pmex, !pmex.isMonitored(), Boolean.FALSE);
                    } else {
                        this.persistenceService.updateMessageExchange(pmex, !pmex.isMonitored(), Boolean.FALSE);
                    }
                }
            }
        } catch (SQLException e) {
            this.log.warning("Failed to store a monitored exchange with Id '" + pmex.getExchangeId() + "'", e);
        } catch (IOException e) {
            this.log.warning("Failed to store a monitored exchange with Id '" + pmex.getExchangeId() + "'", e);
        }
    }

    @Override
    public void exchangeSent(final PetalsMessageExchange exchange) {
        this.log.call("Exchange Id : " + exchange.getExchangeId());

        // this method only makes sense for a message that can be persisted!
        if (!(exchange instanceof PersistedPetalsMessageExchange)) {
            this.log.warning("The message is not a PersistedPetalsMessageExchange: " + exchange.getClass().getName());
            return;
        }

        final PersistedPetalsMessageExchange pmex = (PersistedPetalsMessageExchange) exchange;

        try {
            if (pmex.isMonitored() || pmex.isPersisted()) {
                if (MessageExchange.Role.PROVIDER.equals(pmex.getRole())
                        && pmex.getMessage(PetalsMessageExchange.IN_MSG) != null) {
                    // first time the exchange is add
                    this.persistenceService.addMessageExchange(pmex, !pmex.isMonitored(), Boolean.TRUE);
                } else {
                    this.persistenceService.updateMessageExchange(pmex, !pmex.isMonitored(), Boolean.TRUE);
                }
            }
        } catch (SQLException e) {
            this.log.warning("Failed to store a monitored exchange with Id '" + pmex.getExchangeId() + "'", e);
        } catch (IOException e) {
            this.log.warning("Failed to store a monitored exchange with Id '" + pmex.getExchangeId() + "'", e);
        }
    }

    @Override
    public Map<String, String> getExchange(final String exchangeId) throws MessagingException {
        this.log.call();
        Map<String, String> result = null;
        try {
            final PetalsMessageExchange exchange = this.persistenceService.getMessageExchange(exchangeId);
            if (exchange != null) {
                result = new HashMap<String, String>();
                result.put(Exchange.ID, exchange.getExchangeId());
                result.put(Exchange.PROVIDER_COMPONENT, ((PetalsServiceEndpoint) exchange.getEndpoint())
                        .getLocation().getComponentName());
                result.put(Exchange.PROVIDER_CONTAINER, ((PetalsServiceEndpoint) exchange.getEndpoint())
                        .getLocation().getContainerName());

                final PetalsServiceEndpoint consumerEndpoint = (PetalsServiceEndpoint) exchange.getConsumerEndpoint();
                if (exchange.getConsumerEndpoint() != null) {
                    result.put(Exchange.CONSUMER_COMPONENT, 
                            consumerEndpoint.getLocation().getComponentName());
                    result.put(Exchange.CONSUMER_CONTAINER, 
                            consumerEndpoint.getLocation().getContainerName());
                }

                result.put(Exchange.INTERFACE_NAME, exchange.getInterfaceName().toString());
                result.put(Exchange.SERVICE_NAME, exchange.getService().toString());
                result.put(Exchange.ENDPOINT_NAME, ((PetalsServiceEndpoint) exchange.getEndpoint()).getEndpointName());
                result.put(Exchange.OPERATION_NAME, exchange.getOperation().toString());
                if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
                    result.put(Exchange.ROLE, Role.CONSUMER);
                } else {
                    result.put(Exchange.ROLE, Role.PROVIDER);
                }

                if (MEPPatternConstants.IN_ONLY.equals(exchange.getPattern())) {
                    result.put(Exchange.MEP, MEP.IN_ONLY);
                } else if (MEPPatternConstants.ROBUST_IN_ONLY.equals(exchange.getPattern())) {
                    result.put(Exchange.MEP, MEP.ROBUST_IN_ONLY);
                } else if (MEPPatternConstants.IN_OUT.equals(exchange.getPattern())) {
                    result.put(Exchange.MEP, MEP.IN_OUT);
                } else if (MEPPatternConstants.IN_OPTIONAL_OUT.equals(exchange.getPattern())) {
                    result.put(Exchange.MEP, MEP.IN_OPTIONAL_OUT);
                }

                if (exchange.getStatus() != null) {
                    if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                        result.put(Exchange.STATUS, Status.ACTIVE);
                    } else if (ExchangeStatus.DONE.equals(exchange.getStatus())) {
                        result.put(Exchange.STATUS, Status.DONE);
                    } else {
                        result.put(Exchange.STATUS, Status.ERROR);
                    }
                }

                if (exchange.getError() != null) {
                    result.put(Exchange.EXCEPTION, exchange.getError().getMessage());
                }

                for (Object key : exchange.getPropertyNames()) {
                    result.put(Exchange.PROPERTY_KEY_PREFIX + (String) key, exchange.getProperty(
                            (String) key).toString());
                }

                for (Entry<String, javax.jbi.messaging.NormalizedMessage> entry : exchange.getMessages().entrySet()) {

                    for (Object key : entry.getValue().getPropertyNames()) {
                        result.put(NormalizedMessage.NORMALIZED_MESSAGE_PREFIX + entry.getKey()
                                + "-" + NormalizedMessage.PROPERTY_KEY_PREFIX + (String) key, entry
                                .getValue().getProperty((String) key).toString());
                    }

                    try {
                        result.put(NormalizedMessage.NORMALIZED_MESSAGE_PREFIX + entry.getKey()
                                + "-" + NormalizedMessage.CONTENT, SourceHelper.toString(entry
                                .getValue().getContent()));
                    } catch (TransformerException e) {
                        result.put(NormalizedMessage.CONTENT, e.getMessage());
                    } catch (PoolException e) {
                        result.put(NormalizedMessage.CONTENT, e.getMessage());
                    }

                    for (Object key : entry.getValue().getAttachmentNames()) {
                        result.put(NormalizedMessage.NORMALIZED_MESSAGE_PREFIX + entry.getKey()
                                + "-" + NormalizedMessage.ATTACHMENT_KEY_PREFIX + (String) key,
                                entry.getValue().getAttachment((String) key).toString());
                    }
                }

            }
        } catch (SQLException e) {
            throw new MessagingException(e);
        } catch (IOException e) {
            throw new MessagingException(e);
        } catch (ClassNotFoundException e) {
            throw new MessagingException(e);
        }
        return result;
    }

    @Override
    public Map<String, Long> getExchangeHistory(String exchangeId) throws MessagingException {
        final Map<String, Long> result;

        try {
            result = this.persistenceService.getMessageExchangeHistory(exchangeId);
        } catch (SQLException e) {
            throw new MessagingException(e);
        }

        return result;
    }

    @Override
    public List<String> getExchangeIds(QName interfaceName, QName serviceName, String endpointName,
            QName operationName, long startDate, long endDate) throws MessagingException {
        this.log.call();
        final List<String> result;

        try {
            result = this.persistenceService.getMessageExchangeIds(interfaceName, serviceName,
                    endpointName, operationName, startDate, endDate);
        } catch (SQLException e) {
            throw new MessagingException(e);
        }

        return result;
    }

    @Override
    public List<String> getExchangeIds(String propertyKey, String propertyValue)
            throws MessagingException {
        this.log.call();
        final List<String> result;

        try {
            result = this.persistenceService.getMessageExchangeIds(propertyKey, propertyValue);
        } catch (SQLException e) {
            throw new MessagingException(e);
        } catch (IOException e) {
            throw new MessagingException(e);
        }

        return result;
    }

    @Override
    public int getExchanges(QName interfaceName, QName serviceName, String endpointName,
            QName operationName, long startDate, long endDate) throws MessagingException {
        this.log.call();
        int result;

        try {
            result = this.persistenceService.getMessageExchanges(interfaceName, serviceName,
                    endpointName, operationName, startDate, endDate);
        } catch (SQLException e) {
            throw new MessagingException(e);
        }

        return result;

    }

    @Override
    public List<Map<String, String>> getMonitoredExchanges() {
        this.log.call();

        List<Map<String, String>> result = new ArrayList<Map<String, String>>();

        for (RouterMonitorFilter routerMonitorFilter : this.routerMonitorFilters) {
            Map<String, String> routerMonitor = new HashMap<String, String>();
            if (routerMonitorFilter.getInterfaceName() != null) {
                routerMonitor.put(RouterMonitor.INTERFACE_NAME, routerMonitorFilter
                        .getInterfaceName().toString());
            } else {
                routerMonitor.put(RouterMonitor.INTERFACE_NAME, "null");
            }
            if (routerMonitorFilter.getServiceName() != null) {
                routerMonitor.put(RouterMonitor.SERVICE_NAME, routerMonitorFilter.getServiceName()
                        .toString());
            } else {
                routerMonitor.put(RouterMonitor.SERVICE_NAME, "null");
            }
            routerMonitor.put(RouterMonitor.ENDPOINT_NAME, routerMonitorFilter.getEndpointName());
            if (routerMonitorFilter.getOperationName() != null) {
                routerMonitor.put(RouterMonitor.OPERATION_NAME, routerMonitorFilter
                        .getOperationName().toString());
            } else {
                routerMonitor.put(RouterMonitor.OPERATION_NAME, "null");
            }
            routerMonitor.put(RouterMonitor.STORE_EXCHANGES, Boolean.toString(routerMonitorFilter
                    .isStoreExchanges()));
            result.add(routerMonitor);
        }

        return result;
    }

    @Override
    public long getMonitorStorageDuration() {
        return this.persistenceService.getExchangeStorageDuration();
    }

    @Override
    public void monitorExchanges(QName interfaceName, QName serviceName, String endpointName,
            QName operationName, boolean storeExchanges) throws MessagingException {
        this.log.call();
        final RouterMonitorFilter routerMonitorFilter = new RouterMonitorFilter(interfaceName,
                serviceName, endpointName, operationName, storeExchanges);
        if (this.routerMonitorFilters.contains(routerMonitorFilter)) {
            throw new MessagingException("The Router Monitor filter " + routerMonitorFilter
                    + " already exist");
        } else {
            this.routerMonitorFilters.add(routerMonitorFilter);
        }
    }

    @Override
    public void setMonitorStorageDuration(long duration) {
        this.persistenceService.setExchangeStorageDuration(duration);
    }

    @Override
    public void unmonitorExchanges(QName interfaceName, QName serviceName, String endpointName,
            QName operationName) throws MessagingException {
        final RouterMonitorFilter routerMonitorFilter = new RouterMonitorFilter(interfaceName,
                serviceName, endpointName, operationName, true);

        if (!this.routerMonitorFilters.remove(routerMonitorFilter)) {
            throw new MessagingException("The Router Monitor filter " + routerMonitorFilter
                    + " is not registered");
        }
    }

    /**
     * Start the Fractal component
     * 
     * @throws RoutingException
     */
    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        this.routerMonitorFilters = new HashSet<RouterMonitorFilter>();

        this.log.end();
    }

    /**
     * Stop the Fractal component
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    /**
     * Check if the exchange is monitored.
     * 
     */
    private final void checkMonitored(final PersistedPetalsMessageExchange exchange) {

        // first time the exchange is sent
        for (RouterMonitorFilter routerMonitorFilter : this.routerMonitorFilters) {

            String endpointName = null;
            if (exchange.getEndpoint() != null) {
                endpointName = exchange.getEndpoint().getEndpointName();
            }
            QName interfaceName = exchange.getInterfaceName();
            QName serviceName = exchange.getService();
            QName operationName = exchange.getOperation();

            if ((routerMonitorFilter.getInterfaceName() == null || (interfaceName != null && interfaceName
                    .equals(routerMonitorFilter.getInterfaceName())))
                    && (routerMonitorFilter.getServiceName() == null || (serviceName != null && serviceName
                            .equals(routerMonitorFilter.getServiceName())))
                    && (routerMonitorFilter.getEndpointName() == null || (endpointName != null && endpointName
                            .equals(routerMonitorFilter.getEndpointName())))
                    && (routerMonitorFilter.getOperationName() == null || (operationName != null && operationName
                            .equals(routerMonitorFilter.getOperationName())))) {
                if (routerMonitorFilter.isStoreExchanges()) {
                    exchange.setPersisted(true);
                } else {
                    exchange.setMonitored(true);
                }
                break;
            }
        }
    }
}
