/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.local;

import java.util.logging.Logger;

import javax.jbi.messaging.MessageExchange.Role;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.jbi.messaging.PetalsDeliveryChannel;
import org.ow2.petals.microkernel.transport.AbstractTransporterImpl;
import org.ow2.petals.microkernel.transport.TransportException;
import org.ow2.petals.microkernel.transport.Transporter;
import org.ow2.petals.microkernel.transport.local.monitoring.LocalTransporterMonitoring;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * @author Roland NAUDIN - EBM WebSourcing
 * @author Victor Noel - Linagora
 */
@Component(provides = { @Interface(name = "service", signature = Transporter.class) })
public class LocalTransporterImpl extends AbstractTransporterImpl {

    /**
     * The local transporter monitoring service - Fractal component
     */
    @Requires(name = "localtransportermonitoring")
    protected LocalTransporterMonitoring localTransporterMonitoring;

    public LocalTransporterImpl() {
        super(new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME)));
    }

    @Override
    protected void notifySent(final PetalsMessageExchange exchange) {
        this.log.start();

        if (this.log.isDebugEnabled()) {
            this.log.debug(exchange.getExchangeId()
                    + ": we notify the local transport monitor that the exchange was sent");
        }
        this.localTransporterMonitoring.incMessageSentProbe(exchange);

        super.notifySent(exchange);

        this.log.end();
    }

    @Override
    public void send(final PetalsMessageExchange exchange, final TransportSendContext transportContext)
            throws TransportException {
        this.log.start();

        final String exchangeId = exchange.getExchangeId();

        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending asynchronously exchange " + exchangeId
                    + ", let's see if it is an answer to a send sync or a standalone async send");
        }

        // does the needed checks
        preSend(exchange, transportContext);

        // is that a message that someone is synchronously waiting for?
        // note: we must check the property and not the map, because the entry can be removed from the map
        // when a send times out.
        final boolean answerToSendSync = asBooleanOrFalse(exchange.getProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC));

        // Here we should remove it because we are not waiting for it anymore
        // Note: if it has timed out, it is not in the map anymore...
        // we get it before checking for timeout in order to have nice asserts to detect bugs
        final PetalsMessageExchange oldExchange = removeOfSendSync(exchangeId);

        // there is 3 possible cases:
        // 1) it is a sendsync
        // 1a) and it has timed out
        // 1b) and it hasn't
        // 2) it is not a send sync
        if (answerToSendSync) {
            // case 1)
            if (this.log.isDebugEnabled()) {
                this.log.debug(exchangeId + " is an answer to a send sync");
            }

            // someone is waiting for it: let's release the wait() of the sendSync method
            // (Or it timed out and then we just forget about this answer (as we should according to JBI...))
            synchronized (exchange) {

                // we can clear it now that we know that it came back
                exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, null);

                if (oldExchange == null) {
                    // case 1a)
                    if (this.log.isDebugEnabled()) {
                        this.log.debug(exchangeId
                                + " arrived too late: the send sync already timed out: this answer is ignored then");
                    }
                } else {
                    // case 1b)
                    if (this.log.isDebugEnabled()) {
                        this.log.debug(exchangeId + " has not timed out: we notify the thread waiting for this answer");
                    }
                    this.log.assertOrLog(oldExchange == exchange,
                            "Two different MEX for the same message id " + exchangeId + "in local mode is impossible");
                    // let's wake-up the wait() waiting for this one
                    exchange.notifyAll();
                }
            }
        } else {
            // case 2)
            if (this.log.isDebugEnabled()) {
                this.log.debug(exchangeId + " is a standalone async sent exchange");
            }

            this.log.assertOrLog(oldExchange == null,
                    "This is not an answer for a sync send, but there was an exchange in my pending sync exthange map");

            // in that case, let's give the exchange to the one that should get it
            deliverExchange(exchange);
        }

        // now we can consider the message has been sent
        notifySent(exchange);

        this.log.end();
    }

    @Override
    public PetalsMessageExchange sendSync(final PetalsMessageExchange exchange,
            final TransportSendContext transportContext) throws TransportException, InterruptedException {
        this.log.start();

        // two phases:
        // A) sending the exchange
        // B) waiting for an answer (or timing out)
        // each phase is linked to a role the exchange has, there is two cases:
        // i) sending consumer exchange, receiving provider exchange
        // ii) sending provider exchange, receiving consumer exchange
        // this information helps us reading the logs because of the interleaving of threads happening when we are
        // waiting for an answer and this answer is sent

        final String exchangeId = exchange.getExchangeId();

        // So let's store the current role because we are waiting for an answer in the other role and it will serve as
        // the condition to check to know if we can really start or stop waiting
        final Role preWaitRole = exchange.getRole();

        final String logHint = exchangeId + "/" + preWaitRole;

        // phase A)
        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending synchronously exchange " + logHint
                    + ", let's see if it is an answer to a previous send sync or a new sync send");
        }

        // does the needed checks
        preSendSync(exchange, transportContext);

        // let's see if it an answer to an already send sync message
        final boolean answerToSendSync = asBooleanOrFalse(exchange
                .getProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC));

        final PetalsMessageExchange result;
        synchronized (exchange) {

            if (answerToSendSync) {
                if (this.log.isDebugEnabled()) {
                    this.log.debug(logHint + " is an answer to a send sync: we notify the waiting thread!");
                }

                final PetalsMessageExchange got = getFromSendSync(exchangeId);
                this.log.assertOrLog(got != null,
                        "this is an answer to a sync send, but there is nothing in the map...");
                this.log.assertOrLog(got == exchange,
                        "Two different MEX for the same message id in local mode is not possible");
                // in this case the current send sync is a response of a previous send sync and thus we just need to
                // release it (it is already in the pendingSyncExchanges) so that the first sender can continue its
                // business and then we will wait for an answer
                exchange.notifyAll();
            } else {
                if (this.log.isDebugEnabled()) {
                    this.log.debug(logHint + " is a new send sync");
                }

                // set the property as defined in the JBI specifications 5.5.2.1.3
                // TODO(cdeneux): This property should be used on provider side to avoid to return reply when a timeout
                // occurred on the consumer side. The property value could be the expiration timestamp.
                exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, true);

                // we store it
                storeForSendSync(exchange);

                deliverExchange(exchange);
            }

            // now we can consider the message has been sent
            notifySent(exchange);

            // phase B)

            if (this.log.isDebugEnabled()) {
                this.log.debug("And now we are going to wait for an answer to " + logHint);
            }
            // after that we must keep the lock until the wait to be sure
            // that any notifyAll() does not happen before we start waiting
            // TODO actually, the loop for the wait() first check the condition, so theoretically, we shouldn't even
            // wait in the case a notifyAll happened... but maybe other things need the synchronisation (such as the
            // events like onExchange)

            // So let's wait for an answer
            final boolean timedout = this.waitForAnswerOrTimeout(exchange, preWaitRole);

            // note: we can't just remove the decorator from the map now because we don't know if it wasn't stored in it
            // by the sender (in sync mode) that now released our wait.
            // we thus remove it only in case of timeout or in case of send async of the answer (see send method)

            if (timedout) {
                if (this.log.isDebugEnabled()) {
                    this.log.debug(logHint + " timed out");
                }

                // we remove the exchange from the pending list
                removeOfSendSync(exchangeId);

                // we clean the properties we put
                exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, null);

                this.log.warning("Failed to send synchronously the exchange: " + exchangeId
                        + ": timeout occured");

                result = null;
            } else {
                final Role postWaitRole = exchange.getRole();
                this.log.assertOrLog(postWaitRole != preWaitRole,
                        "After waiting, the role of the exchange hasn't changed");

                final String answerLogHint = exchangeId + "/" + postWaitRole;
                if (this.log.isDebugEnabled()) {
                    this.log.debug(logHint + " was waiting, we got an answer: " + answerLogHint);
                }

                // In local mode, it is the same exchange that is returned
                result = exchange;
            }
        }

        this.log.end();

        return result;
    }
}
