/**
 * Copyright (c) 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.local.monitoring;

import java.util.Collections;
import java.util.Map;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.transport.local.Constants;
import org.ow2.petals.probes.api.KeyedProbesFactory;
import org.ow2.petals.probes.api.KeyedProbesFactoryBuilder;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;
import org.ow2.petals.probes.api.exceptions.ProbeKeyMissingException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.KeyedCounterProbeWithExecutionStatus;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * Local transporter monitoring service implementation.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
@Component(provides = { @Interface(name = "service", signature = LocalTransporterMonitoring.class) })
public class LocalTransporterMonitoringImpl implements LocalTransporterMonitoring {

    /**
     * The logger.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));

    /**
     * The probe counting messages sent by this transporter.
     */
    private final KeyedCounterProbeWithExecutionStatus<ProbeKey> deliveredMessagesCount;

    public LocalTransporterMonitoringImpl() throws MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {
        final KeyedProbesFactoryBuilder<LocalTransporterProbeKey, Long, Long> probesFactoryBuilder = new KeyedProbesFactoryBuilder<LocalTransporterProbeKey, Long, Long>();
        final KeyedProbesFactory<LocalTransporterProbeKey, Long, Long> probesFactory = probesFactoryBuilder
                .getKeyedProbesFactory();
        this.deliveredMessagesCount = probesFactory.createKeyedCounterProbeWithExecutionStatus();
    }

    @Override
    public void incMessageSentProbe(final PetalsMessageExchange exchange) {

        final ProbeKey messageSentProbeKey = new LocalTransporterProbeKey(exchange);
        try {
            this.deliveredMessagesCount.incPending(messageSentProbeKey);
            this.deliveredMessagesCount.move(messageSentProbeKey, ExecutionStatus.SUCCEEDED);
        } catch (ProbeNotStartedException e) {
            this.log.warning("Probe counting asynchronous message exchanges sent is not started. Values of probes could be incorrect.");
        } catch (final ProbeKeyMissingException e) {
            this.log.warning("Probe key missing. Values of probe counting message exchanges could be incorrect.");
        }
    }

    @Override
    public Map<String[], Long> getDeliveredMessagesCount() {
        try {
            return this.deliveredMessagesCount.getConvertedValues();
        } catch (final ProbeNotInitializedException e) {
            this.log.warning("Probe counting message exchanges sent is not initialized. No value available.");
            return Collections.emptyMap();
        }
    }

    @Lifecycle(step = Step.START)
    public void start() throws Exception {
        this.log.start();

        this.deliveredMessagesCount.init();
        this.deliveredMessagesCount.start();

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        try {
            this.deliveredMessagesCount.stop();
            this.deliveredMessagesCount.shutdown();
        } catch (final Exception e) {
            this.log.warning("Can't stop and/or shutdown the counter ", e);
        }

        this.log.end();
    }

}
