/**
 * Copyright (c) 2009-2012 Capgemini, 2009-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.platform.nio;

import java.util.logging.Logger;

import javax.jbi.messaging.MessageExchange.Role;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.jbi.messaging.PetalsDeliveryChannel;
import org.ow2.petals.microkernel.transport.AbstractTransporterImpl;
import org.ow2.petals.microkernel.transport.TransportException;
import org.ow2.petals.microkernel.transport.Transporter;
import org.ow2.petals.microkernel.transport.platform.nio.client.NioClientObjectPool;
import org.ow2.petals.microkernel.transport.platform.nio.selector.NioClient;
import org.ow2.petals.microkernel.transport.platform.nio.server.NioReceiver;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * NIO Implementation of the Transporter.
 * 
 * @author Christophe DENEUX (Capgemini Sud)
 * @author Victor Noel - Linagora
 */
@Component(provides = { @Interface(name = "service", signature = Transporter.class),
        @Interface(name = NioReceiver.NIO_MSG_RECEIVER_ITF, signature = NioReceiver.class) })
public class NioTransporterImpl extends AbstractTransporterImpl implements NioReceiver {

    /**
     * Property stored in the message and used to store the fact that a message was sent in send sync before an answer
     * was sent again.
     * 
     * It is needed because this implementation assumes two nodes communicating over NIO that do not share the map of
     * pending sync messages: if A sendSync a msg to B, it sets the SendSync property, then if B sendSync back to A,
     * then A can't know if the property is here for him, or for B later. A can't use the map to know that (because a
     * timed out message is removed from it) so it needs some extra information to know it.
     * 
     * Note: This property is ONLY set between sending from one NIO transporter node to another NIO transporter node
     * that then unset it, so it isn't visible to anybody else than the NIO transporter implementation.
     */
    private static final String PROPERTY_WAS_SENDSYNC = "org.ow2.petals.microkernel.transport.platform.nio.NioTransporterImpl.WASSENDSYNC";

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * All connections to remote containers are pooled into a keyed-pool. The
     * key is the name of the remote container.
     */
    @Requires(name = NioClientObjectPool.NIO_CLIENT_OBJECT_ITF)
    private NioClientObjectPool clientConnections;

    /**
     * Timeout to send a message exchange.
     */
    private long sendTimeout;

    public NioTransporterImpl() {
        super(new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME)));
    }

    @Override
    public void onMessage(final Object object) {
        this.log.start();

        if (object instanceof PetalsMessageExchange) {

            try {
                this.checkTransporterState();
            } catch (final TransportException e) {
                this.log.warning("Received a message but the transporter is stopped. Data are discarded.", e);
                this.log.end();
                return;
            }

            final PetalsMessageExchange messageExchange = (PetalsMessageExchange) object;

            final String exchangeId = messageExchange.getExchangeId();

            if (this.log.isDebugEnabled()) {
                this.log.debug("Receiving exchange " + exchangeId
                        + " on NIO, let's see if it is an answer to a send sync or a standalone async send");
            }

            // we need to check that information from the received message
            // wsasendsync means that we sent it in send sync, and that on the other side, they stored the information
            // for us
            final boolean answerToSendSync = asBooleanOrFalse(messageExchange.getProperty(PROPERTY_WAS_SENDSYNC));
            // we can now reset it to nothing without touching to the original SendSync property
            messageExchange.setProperty(PROPERTY_WAS_SENDSYNC, null);

            // we don't remove it from sendSync because it is needed for timeout checking
            final PetalsMessageExchange oldExchange = getFromSendSync(exchangeId);

            // there is 3 possible cases:
            // 1) it is a sendsync
            // 1a) and it has timed out
            // 1b) and it hasn't
            // 2) it is not a send sync
            if (answerToSendSync) {
                // case 1)
                if (this.log.isDebugEnabled()) {
                    this.log.debug(exchangeId + " is an answer to a send sync");
                }

                // someone is waiting for it, let's release the wait() of the sendSync method
                // (or it timed out and then we forget about this answer as we should according to JBI)

                if (oldExchange == null) {
                    // case 1a)
                    if (this.log.isDebugEnabled()) {
                        this.log.debug(exchangeId
                                + " arrived too late: the send sync already timed out: this answer is ignored then");
                    }
                } else {
                    // case 1b)
                    if (this.log.isDebugEnabled()) {
                        this.log.debug(exchangeId + " has not timed out: we notify the thread waiting for this answer");
                    }

                    // we must synchronize on the old one, this is the one we know about!!!
                    synchronized (oldExchange) {
                        // let's put the new one instead
                        replaceInSendSync(messageExchange);
                        // let's wake-up the wait() waiting for this one
                        oldExchange.notifyAll();
                    }
                }
            } else {
                // case 2)
                if (this.log.isDebugEnabled()) {
                    this.log.debug(exchangeId + " is a standalone async sent exchange");
                }

                this.log.assertOrLog(oldExchange == null,
                        "This is not an answer for a sync send, but there was an exchange in my pending sync exthange map");

                deliverExchange(messageExchange);
            }
        } else {
            // This case shouldn't happen...
            this.log.warning("Unexpected received data (shouldn't happen). Data is discarded.");
        }

        this.log.end();
    }

    private void nioSend(final PetalsMessageExchange exchange, final TransportSendContext transportContext,
            final boolean sendSync) throws TransportException {
        this.log.start();

        final String containerName = transportContext.destination.getContainerName();

        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending over NIO exchange " + exchange.getExchangeId() + " to " + containerName);
        }

        // first let's store the previous value (set by the previous sender to which this send must then be
        // an answer) so that when it is received, he still knows about it.
        exchange.setProperty(PROPERTY_WAS_SENDSYNC, exchange.getProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC));

        // set (or unset) the property as defined in the JBI specifications 5.5.2.1.3
        // TODO(cdeneux): This property should be used on provider side to avoid to return reply when a timeout
        // occurred on the consumer side. The property value could be the expiration timestamp.
        exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, sendSync ? true : null);

        try {
            final NioClient clientConnection = this.clientConnections.borrowObject(containerName);
            try {
                clientConnection.send(exchange, this.sendTimeout);

                // let's notify that the message was sent (only if no error happened)
                notifySent(exchange);

            } finally {
                this.clientConnections.returnObject(containerName, clientConnection);
            }
        } catch (final Exception e) {
            throw new TransportException(e);
        }

    }

    /**
     * This method does following action :
     * <ol>
     * <li>Serialize the message exchange.</li>
     * <li>Get the destination of the message.</li>
     * <li>Add it to the message.</li>
     * <li>Send it to the transport agent.</li>
     * </ol>
     */
    @Override
    public void send(final PetalsMessageExchange exchange, final TransportSendContext transportContext)
            throws TransportException {
        this.log.start();

        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending asynchronously exchange " + exchange.getExchangeId());
        }

        // does the needed checks
        preSend(exchange, transportContext);

        // we then send it over NIO
        this.nioSend(exchange, transportContext, false);

        this.log.end();
    }

    @Override
    public PetalsMessageExchange sendSync(final PetalsMessageExchange exchange,
            final TransportSendContext transportContext) throws TransportException, InterruptedException {
        this.log.start();

        // two phases:
        // A) sending the exchange
        // B) waiting for an answer (or timing out)
        // each phase is linked to a role the exchange has, there is two cases:
        // i) sending consumer exchange, receiving provider exchange
        // ii) sending provider exchange, receiving consumer exchange
        // this information helps us reading the logs because of the interleaving of threads happening when we are
        // waiting for an answer and this answer is sent

        final String exchangeId = exchange.getExchangeId();

        // So let's store the current role because we are waiting for an answer in the other role and it will serve as
        // the condition to check to know if we can really start or stop waiting
        final Role preWaitRole = exchange.getRole();

        final String logHint = exchangeId + "/" + preWaitRole;

        // phase A)
        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending synchronously exchange " + logHint);
        }

        // does the needed checks
        preSendSync(exchange, transportContext);

        final PetalsMessageExchange answer;
        synchronized (exchange) {

            // we need to store it so that we can find it back in onMessage when receiving an answer
            storeForSendSync(exchange);

            // we then send it over NIO
            this.nioSend(exchange, transportContext, true);

            // after that we must keep the lock until the wait to be sure
            // that any notifyAll() from an answer does not happen before we start waiting
            // TODO actually, the loop for the wait() first check the condition, so theoretically, we shouldn't even
            // wait in the case a notifyAll happened... but maybe other things need the synchronisation (such as the
            // events in nioSend)

            // phase B)
            if (this.log.isDebugEnabled()) {
                this.log.debug("And now we are going to wait for an answer to " + logHint);
            }

            // and wait for an answer
            final boolean timedout = this.waitForAnswerOrTimeout(exchange, preWaitRole);

            // we remove the exchange from the pending list now that we got the answer
            // the idea is that if it timedout, we get back the original one, and if not, we get back an answer
            // note: this is different than in local mode: there we keep the exchange in the map since it is in local
            // and a sendsync, but here, the map is on another node for the answer
            final PetalsMessageExchange oldExchange = removeOfSendSync(exchange.getExchangeId());

            if (timedout) {
                if (this.log.isDebugEnabled()) {
                    this.log.debug(logHint + " timed out");
                }

                this.log.assertOrLog(oldExchange == exchange,
                        "sendSync timed out but I got an answer different than the one originally stored in my map.");

                // we clean the properties we put
                oldExchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, null);
                oldExchange.setProperty(PROPERTY_WAS_SENDSYNC, null);

                this.log.warning("Failed to send synchronously the exchange: " + oldExchange.getExchangeId()
                        + ": timeout occured");

                answer = null;
            } else {
                this.log.assertOrLog(oldExchange != exchange,
                        "After waiting, the answer I got is the original exchange");

                final Role postWaitRole = oldExchange.getRole();
                this.log.assertOrLog(postWaitRole != preWaitRole,
                        "After waiting, the role of the exchange hasn't changed");

                final String answerLogHint = exchangeId + "/" + postWaitRole;
                if (this.log.isDebugEnabled()) {
                    this.log.debug(logHint + " was waiting, we got an answer: " + answerLogHint);
                }

                answer = oldExchange;
            }
        }

        this.log.end();

        return answer;
    }

    @Lifecycle(step = Step.START)
    @Override
    public void start() {
        this.log.start();
        super.start();

        this.sendTimeout = this.configurationService.getContainerConfiguration().getTCPSendTimeout();

        this.log.end();
    }
}