/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InvalidSubdomainConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoSubdomainConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.SubdomainAlreadyExistException;

/**
 * Unit tests of {@link ConfigurationServiceImpl#addSubdomainConfiguration(SubDomainConfiguration)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_AddSubdomainConfigurationTest extends AbstractConfigurationServiceImplTest {

    private final static String SUBDOMAIN_NAME_1 = "subdomain-1";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the add of a sub-domain configuration against:
     * <ul>
     * <li>no sub-domain configuration is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because no sub-domain configuration is provided.
     * </p>
     */
    @Test(expected = NoSubdomainConfigurationProvidedException.class)
    public void testNoSubdomainProvided() throws ConfigurationException, InvocationTargetException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.addSubdomainConfiguration(null);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the add of a sub-domain configuration against:
     * <ul>
     * <li>the sub-domain configuration provided has a name unset (ie. <code>null</code>)</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the sub-domain configuration is invalid.
     * </p>
     */
    @Test(expected = InvalidSubdomainConfigurationProvidedException.class)
    public void testSubdomainProvidedHasANullName() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final SubDomainConfiguration subdomainToAdd = new SubDomainConfiguration();
            cfgSvc.addSubdomainConfiguration(subdomainToAdd);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the add of a sub-domain configuration against:
     * <ul>
     * <li>the sub-domain configuration provided is the current sub-domain configuration</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the sub-domain configuration is invalid because it is the configuration
     * of the current sub-domain.
     * </p>
     */
    @Test(expected = InvalidSubdomainConfigurationProvidedException.class)
    public void testSubdomainProvidedIsTheCurrentOne() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final SubDomainConfiguration subdomainToAdd = new SubDomainConfiguration();
            subdomainToAdd.setName(ConfigurationTestUtils.SUBDOMAIN_NAME_0);
            cfgSvc.addSubdomainConfiguration(subdomainToAdd);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the add of a sub-domain configuration against:
     * <ul>
     * <li>the topology contains 1 sub-domain</li>
     * <li>a sub-domain is added with a valid sub-domain configuration</li>
     * <li>the sub-domain is added again</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>On the first add:
     * <ul>
     * <li>No error occurs</li>
     * <li>The sub-domain is correctly added</li>
     * <li>The topology contains 2 sub-domains</li>
     * </ul>
     * </li>
     * <li>An error occurs on the second add because the sub-domain already exist</li>
     * </p>
     */
    @Test
    public void testSubdomainProvidedIsValid() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final SubDomainConfiguration subdomainToAdd = new SubDomainConfiguration();
            subdomainToAdd.setName(SUBDOMAIN_NAME_1);
            cfgSvc.addSubdomainConfiguration(subdomainToAdd);

            assertNotNull("sub-domain 0 no more exists",
                    cfgSvc.getSubDomainConfiguration(ConfigurationTestUtils.SUBDOMAIN_NAME_0));
            assertNotNull("sub-domain 1 does not exist", cfgSvc.getSubDomainConfiguration(SUBDOMAIN_NAME_1));
            assertEquals("unexpected sub-domain number", cfgSvc.getSubDomainsConfiguration().size(), 2);

            try {
                cfgSvc.addSubdomainConfiguration(subdomainToAdd);
                fail("Exception " + SubdomainAlreadyExistException.class.getName() + " not thrown");
            } catch (final SubdomainAlreadyExistException e) {
                assertEquals("unexpected already existing sub-domain", SUBDOMAIN_NAME_1,
                        e.getAlreadyExistingSubdomainName());
            }

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
