/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerDoesNotExistException;
import org.ow2.petals.microkernel.api.configuration.exception.InvalidSubdomainConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoSubdomainConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.SubdomainDoesNotExistException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.SubdomainMode;
import org.ow2.petals.topology.generated.TransportService;

/**
 * Unit tests of {@link ConfigurationServiceImpl#removeSubdomainConfiguration(SubDomainConfiguration)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_RemoveSubdomainConfigurationTest extends AbstractConfigurationServiceImplTest {

    protected final static String SUBDOMAIN_NAME_1 = "subdomain-1";

    protected final static SubdomainMode SUBDOMAIN_MODE_1 = SubdomainMode.STATIC;

    protected final static String SUBDOMAIN_DESCRIPTION_1 = "A sample sub-domain 1 configuration";

    private final static String CONTAINER_NAME_1 = "sample-1";

    private final static String CONTAINER_DESCRIPTION_1 = "description of the container 1";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    public ConfigurationServiceImpl_RemoveSubdomainConfigurationTest() {
        super();

        final Domain domain = this.topology.getDomain();
        final Subdomain subdomain1 = new Subdomain();
        domain.getSubDomain().add(subdomain1);
        subdomain1.setName(SUBDOMAIN_NAME_1);
        subdomain1.setMode(SUBDOMAIN_MODE_1);
        subdomain1.setDescription(SUBDOMAIN_DESCRIPTION_1);

        final Container container1 = new Container();
        subdomain1.getContainer().add(container1);
        container1.setName(CONTAINER_NAME_1);
        container1.setDescription(CONTAINER_DESCRIPTION_1);
        container1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        container1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        container1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

        final JmxService jmxService1 = new JmxService();
        container1.setJmxService(jmxService1);

        final TransportService transportService1 = new TransportService();
        transportService1.setTcpPort(7781);
        container1.setTransportService(transportService1);
    }

    /**
     * <p>
     * Check the deletion of a sub-domain configuration against:
     * <ul>
     * <li>no sub-domain configuration is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because no sub-domain configuration is provided.
     * </p>
     */
    @Test(expected = NoSubdomainConfigurationProvidedException.class)
    public void testNoSubdomainProvided() throws ConfigurationException, InvocationTargetException {
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.removeSubdomainConfiguration(null);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the deletion of a sub-domain configuration against:
     * <ul>
     * <li>the sub-domain configuration provided has a name unset (ie. <code>null</code>)</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the sub-domain configuration is unknown.
     * </p>
     */
    @Test(expected = SubdomainDoesNotExistException.class)
    public void testSubdomainProvidedHasANullName() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final SubDomainConfiguration containerToRemove = new SubDomainConfiguration();
            cfgSvc.removeSubdomainConfiguration(containerToRemove);
        } catch (final ContainerDoesNotExistException e) {
            assertEquals("Unexpected container name", e.getUnexistingContainerName(), null);
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the deletion of a sub-domain configuration against:
     * <ul>
     * <li>the sub-domain configuration provided is unknown</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the sub-domain configuration is unknown.
     * </p>
     */
    @Test(expected = SubdomainDoesNotExistException.class)
    public void testSubdomainProvidedIsUnknown() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final String subdomainNameToRemove = "unknown-subdomain";
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final SubDomainConfiguration subdomainToRemove = new SubDomainConfiguration();
            subdomainToRemove.setName(subdomainNameToRemove);
            cfgSvc.removeSubdomainConfiguration(subdomainToRemove);
        } catch (final SubdomainDoesNotExistException e) {
            assertEquals("Unexpected sub-domain name", e.getUnexistingSubdomainName(), subdomainNameToRemove);

            assertNotNull("sub-domain 0 no more exists",
                    cfgSvc.getSubDomainConfiguration(ConfigurationTestUtils.SUBDOMAIN_NAME_0));
            assertNotNull("sub-domain 1 no more exists", cfgSvc.getSubDomainConfiguration(SUBDOMAIN_NAME_1));
            assertEquals("unexpected sub-domain number", cfgSvc.getSubDomainsConfiguration().size(), 2);
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the deletion of a sub-domain configuration against:
     * <ul>
     * <li>the sub-domain configuration provided is associated to the current sub-domain</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the current sub-domain configuration cannot be removed.
     * </p>
     */
    @Test(expected = InvalidSubdomainConfigurationProvidedException.class)
    public void testSubdomainProvidedIsCurrent() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final SubDomainConfiguration subdomainToRemove = new SubDomainConfiguration();
            subdomainToRemove.setName(ConfigurationTestUtils.SUBDOMAIN_NAME_0);
            cfgSvc.removeSubdomainConfiguration(subdomainToRemove);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the deletion of a sub-domain configuration against:
     * <ul>
     * <li>the topology contains 2 sub-domains</li>
     * <li>the sub-domain configuration provided is valid</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>
     * <li>The sub-domain is correctly removed</li>
     * <li>Containers of the sub-domain are removed</li>
     * <li>Only one sub-domain remains inside the topology</li>
     * </p>
     */
    @Test
    public void testSubdomainProvidedIsKnown() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final SubDomainConfiguration subdomainToRemove = new SubDomainConfiguration();
            subdomainToRemove.setName(SUBDOMAIN_NAME_1);
            cfgSvc.removeSubdomainConfiguration(subdomainToRemove);

            assertNotNull("sub-domain 0 no more exists",
                    cfgSvc.getSubDomainConfiguration(ConfigurationTestUtils.SUBDOMAIN_NAME_0));
            assertNull("sub-domain 1 exists always", cfgSvc.getSubDomainConfiguration(SUBDOMAIN_NAME_1));
            assertEquals("unexpected sub-domain number", 1, cfgSvc.getSubDomainsConfiguration().size());
            assertNotNull("container 0 no more exists",
                    cfgSvc.getContainerConfiguration(ConfigurationTestUtils.CONTAINER_NAME_0));
            assertNull("container 1 exists always", cfgSvc.getContainerConfiguration(CONTAINER_NAME_1));
            assertEquals("unexpected container number", 1, cfgSvc.getContainersConfiguration().size());

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
