/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.exchange;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.net.URI;
import java.util.logging.Logger;

import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.RobustInOnly;
import javax.xml.namespace.QName;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeCheckHelper;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;

import com.ebmwebsourcing.easycommons.thread.ExecutionContext;

/**
 * @author Olivier Fabre - EBM WebSourcing
 */
public class PersistedMessageExchangeFactoryImplTest extends MessageExchangeCheckHelper {

    protected final PetalsServiceEndpoint consumerEndpointMock = Mockito.mock(PetalsServiceEndpoint.class);

    protected final Logger logger = Logger.getLogger("testLogger");

    protected MessageExchangeFactory messageExchangeFactory;

    protected MessageExchangeFactory createTestMessageExchangeFactory() {
        return PersistedMessageExchangeFactoryImpl.create(
                consumerEndpointMock, logger);
    }

    @Before
    public final void before() {
        messageExchangeFactory = createTestMessageExchangeFactory();
        PetalsExecutionContext.putContainerName("containerName");
    }

    @After
    public final void after() {
        ExecutionContext.getProperties().clear();
    }

    @Test
    public void testCreateExchangeFromServAndOp() throws MessagingException {
        QName serviceName = new QName("targetService");
        QName opName = new QName("targetOperation");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(serviceName,
                opName);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
        assertGeneratedExchangeUseProvidedServiceAndOp(actualMessageExchange, serviceName, opName);
    }

    @Test
    public void testCreateExchangeFromLegacyInOnlyPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2004/08/wsdl/in-only");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromLegacy2InOnlyPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2006/01/wsdl/in-only");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromNewInOnlyPattern() throws MessagingException {
        URI pattern = AbsItfOperation.MEPPatternConstants.IN_ONLY.value();

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromLegacyRobustInOnlyPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2004/08/wsdl/robust-in-only");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsRobustInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromLegacy2RobustInOnlyPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2006/01/wsdl/robust-in-only");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsRobustInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromNewRobustInOnlyPattern() throws MessagingException {
        URI pattern = AbsItfOperation.MEPPatternConstants.ROBUST_IN_ONLY.value();

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsRobustInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromLegacyInOptOutPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2004/08/wsdl/in-opt-out");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOptOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromLegacy2InOptOutPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2006/01/wsdl/in-opt-out");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOptOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromNewInOptOutPattern() throws MessagingException {
        URI pattern = AbsItfOperation.MEPPatternConstants.IN_OPTIONAL_OUT.value();

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOptOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromLegacyInOutPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2004/08/wsdl/in-out");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromLegacy2InOutPattern() throws Exception {
        URI pattern = new URI("http://www.w3.org/2006/01/wsdl/in-out");

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromNewInOutPattern() throws MessagingException {
        URI pattern = AbsItfOperation.MEPPatternConstants.IN_OUT.value();

        MessageExchange actualMessageExchange = messageExchangeFactory.createExchange(pattern);

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateExchangeFromNonJBIPattern() {
        URI pattern = URI.create("http://non/jbi/pattern");

        try {
            messageExchangeFactory.createExchange(pattern);
            fail("A MessagingException must be thrown");
        } catch (MessagingException e) {
            // must thow an exception, its ok
        }

    }

    @Test
    public void testCreateInOnlyExchange() throws MessagingException {
        MessageExchange actualMessageExchange = messageExchangeFactory.createInOnlyExchange();

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateInOptionalOutExchange() throws MessagingException {
        MessageExchange actualMessageExchange = messageExchangeFactory
                .createInOptionalOutExchange();

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOptOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateInOutExchange() throws MessagingException {
        MessageExchange actualMessageExchange = messageExchangeFactory.createInOutExchange();

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsInOut(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    @Test
    public void testCreateRobustInOnlyExchange() throws MessagingException {
        MessageExchange actualMessageExchange = messageExchangeFactory.createRobustInOnlyExchange();

        assertGeneratedExchangeIsAMessageExchangeDecorator(actualMessageExchange);
        assertGeneratedExchangeIsRobustInOnly(actualMessageExchange);
        assertIdIsPetalsQualifiedUUID(actualMessageExchange);
    }

    private void assertGeneratedExchangeIsAMessageExchangeDecorator(
            MessageExchange actualMessageExchange) {
        assertTrue("Returned message exchange isn't a MessageExchangeDecorator",
                actualMessageExchange instanceof MessageExchangeWrapper);
    }

    private void assertGeneratedExchangeUseProvidedServiceAndOp(
            MessageExchange actualMessageExchange, QName serviceName, QName opName) {
        assertEquals("Bad service on created exchange", serviceName,
                actualMessageExchange.getService());
        assertEquals("Bad operation on created exchange", opName,
                actualMessageExchange.getOperation());
    }

    private void assertGeneratedExchangeIsInOnly(MessageExchange actualMessageExchange) {
        assertTrue("Bad pattern on created exchange expected 'InOnly' was '"
                + actualMessageExchange.getClass().getName() + "'",
                actualMessageExchange instanceof InOnly);
    }

    private void assertGeneratedExchangeIsRobustInOnly(MessageExchange actualMessageExchange) {
        assertTrue("Bad pattern on created exchange expected 'RobustInOnly' was '"
                + actualMessageExchange.getClass().getName() + "'",
                actualMessageExchange instanceof RobustInOnly);
    }

    private void assertGeneratedExchangeIsInOptOut(MessageExchange actualMessageExchange) {
        assertTrue("Bad pattern on created exchange expected 'InOptOut' was '"
                + actualMessageExchange.getClass().getName() + "'",
                actualMessageExchange instanceof InOptionalOut);
    }

    private void assertGeneratedExchangeIsInOut(MessageExchange actualMessageExchange) {
        assertTrue("Bad pattern on created exchange expected 'InOut' was '"
                + actualMessageExchange.getClass().getName() + "'",
                actualMessageExchange instanceof InOut);
    }

}
